package cn.com.qmth.stmms.ms.marking.service;

import cn.com.qmth.stmms.ms.commons.config.SystemConfig;
import cn.com.qmth.stmms.ms.core.domain.CheckData;
import cn.com.qmth.stmms.ms.core.domain.Level;
import cn.com.qmth.stmms.ms.core.domain.MarkSubject;
import cn.com.qmth.stmms.ms.core.domain.Paper;
import cn.com.qmth.stmms.ms.core.domain.enums.CheckDataStatusEnum;
import cn.com.qmth.stmms.ms.core.domain.enums.CheckItemEnum;
import cn.com.qmth.stmms.ms.core.repository.CheckDataRepo;
import cn.com.qmth.stmms.ms.core.repository.LevelRepo;
import cn.com.qmth.stmms.ms.core.repository.MarkSubjectRepo;
import cn.com.qmth.stmms.ms.core.repository.PaperRepo;
import com.alibaba.fastjson.JSONObject;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.io.File;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

@Service
public class CheckDataService {

    private static final String SEPARATE = "#";

    @Resource
    SystemConfig systemConfig;

    @Resource
    CheckDataRepo checkDataRepo;

    @Resource
    MarkSubjectRepo markSubjectRepo;

    @Resource
    PaperRepo paperRepo;

    @Resource
    LevelRepo levelRepo;

    /**
     * 异步任务做检查
     *
     * @param checkData 任务对象
     */
    @Async
    public void syncCheckData(CheckData checkData) {
        // 执行中
        checkData.setStatus(CheckDataStatusEnum.RUNNING);
        checkDataRepo.saveAndFlush(checkData);

        int errorCount = 0;
        List<JSONObject> list = null;
        try {
            Long workId = checkData.getWorkId();
            if (CheckItemEnum.PAPER_SIZE.equals(checkData.getCheckItem())) {
                errorCount = checkPaperSize(workId);
            } else if (CheckItemEnum.ALL_SCORE.equals(checkData.getCheckItem())) {
                errorCount = checkAllScore(workId);
            } else if (CheckItemEnum.LEVEL_SCORE_MATCH.equals(checkData.getCheckItem())) {
                list = checkLevelScoreMatch(workId);
            } else if (CheckItemEnum.LEVEL_DIFF.equals(checkData.getCheckItem())) {
                list = checkLevelDiff(workId, checkData.getParamValue());
            } else if (CheckItemEnum.SCORE_DIFF.equals(checkData.getCheckItem())) {
                list = checkScoreDiff(workId, checkData.getParamValue());
            } else {
                throw new Exception("没有此类型任务");
            }
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            checkData.setStatus(CheckDataStatusEnum.FINISH);
            checkData.setResult(errorCount == 0);
            checkData.setErrorCount(errorCount);
            if (list != null && list.size() > 0) {
                checkData.setContent(JSONObject.toJSONString(list));
            }
            checkData.setEndTime(new Date());
            checkDataRepo.saveAndFlush(checkData);
        }
    }

    /**
     * 检查图片大小0KB
     *
     * @param workId 工作ID
     */
    private int checkPaperSize(Long workId) {
        int count = 0;
        // image
        File imageDir = new File(systemConfig.getImageDir() + File.separator + workId);
        count = getZeroCount(count, imageDir);

        // thumb
        File thumbDir = new File(systemConfig.getThumbDir() + File.separator + workId);
        count = getZeroCount(count, thumbDir);

        // sheet
        File sheetDir = new File(systemConfig.getSheetDir() + File.separator + workId);
        count = getZeroCount(count, sheetDir);
        return count;
    }

    private int getZeroCount(int count, File fileDir) {
        List<File> files = getAllFile(fileDir);
        if (files != null && files.size() > 0) {
            int zeroCount = Math.toIntExact(files.stream().filter(m -> m.length() == 0).count());
            count = count + zeroCount;
        }
        return count;
    }

    /**
     * 校验每张试卷都有分数
     *
     * @param workId 工作ID
     */
    private int checkAllScore(Long workId) {
        List<Paper> paperList = paperRepo.findByWorkIdAndIsMissing(workId, false);
        if (paperList != null && paperList.size() > 0) {
            return Math.toIntExact(paperList.stream().filter(m -> m.getScore() == null).count());
        }
        return 0;
    }

    /**
     * 每张试卷分数都在其所属档位分数区间内
     *
     * @param workId 工作ID
     */
    private List<JSONObject> checkLevelScoreMatch(Long workId) {
        List<Paper> paperList = paperRepo.findByWorkIdAndIsMissing(workId, false);
        List<JSONObject> list = new ArrayList<>();
        if (paperList != null && paperList.size() > 0) {
            // 有未打分的，算未匹配
            List<Paper> papers = paperList.stream().filter(m -> m.getScore() == null).collect(Collectors.toList());
            if (papers.size() > 0) {
                list = papers.stream().map(p -> {
                    JSONObject object = new JSONObject();
                    object.put("examNumber", p.getExamNumber());
                    object.put("studentName", p.getStudentName());
                    return object;
                }).collect(Collectors.toList());
            } else {
                //分档、打分完成后，做匹配校验
                List<Level> levels = levelRepo.findByWorkId(workId);
                Map<String, Level> levelMap = levels.stream().collect(Collectors.toMap(Level::getCode, Function.identity()));

                list = paperList.stream().filter(p -> {
                    // 未分档的，不参与统计
                    if (p.getLevel() == null) {
                        return false;
                    }
                    Level level = levelMap.get(p.getLevel());
                    // 正常范围，是左开右闭
                    return p.getScore().intValue() <= level.getMinScore() || p.getScore().intValue() >= level.getMaxScore();
                }).map(p -> {
                    JSONObject object = new JSONObject();
                    object.put("examNumber", p.getExamNumber());
                    object.put("studentName", p.getStudentName());
                    return object;
                }).collect(Collectors.toList());
            }
        }
        return list;
    }

    /**
     * 同一考生各科目档位差
     *
     * @param workId     工作ID
     * @param paramValue 变量参数值
     */
    private List<JSONObject> checkLevelDiff(Long workId, int paramValue) {
        List<MarkSubject> markSubjects = markSubjectRepo.findAllByWorkIdAndEnableTrue(workId);

        List<Level> levels = levelRepo.findByWorkId(workId);
        Map<String, Level> levelMap = levels.stream().collect(Collectors.toMap(Level::getCode, Function.identity()));

        List<Paper> paperList = paperRepo.findByWorkIdAndIsMissing(workId, false);
        Map<String, List<Paper>> collect = paperList.stream().collect(Collectors.groupingBy(m -> m.getExamNumber() + SEPARATE + m.getStudentName()));
        List<JSONObject> list = new ArrayList<>();
        for (Map.Entry<String, List<Paper>> entry : collect.entrySet()) {
            String key = entry.getKey();
            List<Paper> papers = entry.getValue();
            // 只考部分科目，校验不通过
            if (papers.size() != markSubjects.size()) {
                list.add(assembleErrorData(key, "只考" + papers.size() + "科"));
            } else {
                long noLevel = papers.stream().filter(m -> m.getLevel() == null).count();
                if (noLevel > 0) {
                    list.add(assembleErrorData(key, "有" + noLevel + "科未分档"));
                } else {
                    List<Integer> integers = papers.stream().map(m -> levelMap.get(m.getLevel()).getLevelValue()).collect(Collectors.toList());
                    // true为校验通过，false为不通过
                    boolean checkFlag = calcDiff(integers, paramValue);
                    if (!checkFlag) {
                        list.add(assembleErrorData(key, "最大差值大于" + paramValue));
                    }
                }
            }
        }
        return list;
    }

    /**
     * 同一考生各科目成绩差
     *
     * @param workId     工作ID
     * @param paramValue 变量参数值
     */
    private List<JSONObject> checkScoreDiff(Long workId, int paramValue) {
        List<MarkSubject> markSubjects = markSubjectRepo.findAllByWorkIdAndEnableTrue(workId);
        List<Paper> paperList = paperRepo.findByWorkIdAndIsMissing(workId, false);
        Map<String, List<Paper>> collect = paperList.stream().collect(Collectors.groupingBy(m -> m.getExamNumber() + SEPARATE + m.getStudentName()));
        List<JSONObject> list = new ArrayList<>();
        for (Map.Entry<String, List<Paper>> entry : collect.entrySet()) {
            String key = entry.getKey();
            List<Paper> papers = entry.getValue();
            // 只考部分科目，校验不通过
            if (papers.size() != markSubjects.size()) {
                list.add(assembleErrorData(key, "只考" + papers.size() + "科"));
            } else {
                long noScore = papers.stream().filter(m -> m.getLevel() != null && m.getScore() == null).count();
                if (noScore > 0) {
                    list.add(assembleErrorData(key, "有" + noScore + "科已分档未打分"));
                } else {
                    List<Integer> integers = papers.stream().map(m -> m.getScore().intValue()).collect(Collectors.toList());
                    // true为校验通过，false为不通过
                    boolean checkFlag = calcDiff(integers, paramValue);
                    if (!checkFlag) {
                        list.add(assembleErrorData(key, "最大差值大于" + paramValue));
                    }
                }
            }
        }
        return list;
    }

    /**
     * 计算差值
     *
     * @param integers   值集合
     * @param paramValue 变量
     * @return true为通过，false为不通过
     */
    private boolean calcDiff(List<Integer> integers, int paramValue) {
        List<Integer> diffList = new ArrayList<>();
        for (int i = 0; i < integers.size(); i++) {
            for (int j = i; j < integers.size(); j++) {
                if (i != j) {
                    int diffValue = Math.abs(integers.get(i) - integers.get(j));
                    diffList.add(diffValue);
                }
            }
        }
        // A和C，差值为2，若paramValue=2,不通过，返回false,paramValue = 1,通过，返回true
        int maxDiff = diffList.stream().max(Comparator.comparing(i -> i)).get();
        return maxDiff < paramValue;
    }

    /**
     * 组装错误信息
     *
     * @param key          分组key(准考证号#考生姓名)
     * @param errorMessage 错误信息
     */
    private JSONObject assembleErrorData(String key, String errorMessage) {
        JSONObject object = new JSONObject();
        String[] strings = key.split(SEPARATE);
        object.put("examNumber", strings[0]);
        object.put("studentName", strings[1]);
        object.put("errorMessage", errorMessage);
        return object;
    }

    /**
     * 获取指定文件夹下所有文件，不含文件夹
     *
     * @param dirFile 文件夹
     */
    public static List<File> getAllFile(File dirFile) {
        // 如果文件夹不存在或着不是文件夹，则返回 null
        if (Objects.isNull(dirFile) || !dirFile.exists() || dirFile.isFile())
            return null;

        File[] childrenFiles = dirFile.listFiles();
        if (Objects.isNull(childrenFiles) || childrenFiles.length == 0)
            return null;

        List<File> files = new ArrayList<>();
        for (File childFile : childrenFiles) {

            // 如果时文件，直接添加到结果集合
            if (childFile.isFile()) {
                files.add(childFile);
            } else {
                // 如果是文件夹，则将其内部文件添加进结果集合
                List<File> cFiles = getAllFile(childFile);
                if (Objects.isNull(cFiles) || cFiles.isEmpty()) {
                    continue;
                }
                files.addAll(cFiles);
            }
        }
        return files;
    }
}
